/*
Copyright 2015 Container Solutions

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package main

import (
	"github.com/hashicorp/terraform/helper/schema"
	"github.com/hashicorp/terraform/plugin"
	"github.com/hashicorp/terraform/terraform"
	"net/http"
        "encoding/xml"
        "fmt"
        "log"
        "io/ioutil"
        "strings"
        "crypto/rand"
	"time"
)

type CloudBrokerClient struct {
	Username     string
	Password   string
        Endpoint   string
	Timeout    int
	MaxRetries int
}

type InstancePost struct {
        XMLName    xml.Name `xml:"instance"`
        Id string `xml:"id"`
        Name string `xml:"name"`
        SoftwareId string `xml:"software_id"`
        ResourceId string `xml:"resource_id"`
        RegionId string `xml:"region_id"`
        InstanceTypeId string `xml:"instance_type_id"`
        Isolated string `xml:"isolated"`
        KeyPairId string `xml:"key_pair_id"`
        DisableAutostop string `xml:"disable_autostop"`
        OpenedPort string `xml:"opened_port"`
}

type InstanceGet struct {
        XMLName    xml.Name `xml:"instance"`
        Id string `xml:"id"`
        Name string `xml:"name"`
        Status string `xml:"status"`
        SoftwareId string `xml:"software-id"`
        ResourceId string `xml:"resource-id"`
        RegionId string `xml:"region-id"`
        InstanceTypeId string `xml:"instance-type-id"`
        Isolated string `xml:"isolated"`
        KeyPairId string `xml:"key-pair-id"`
        DisableAutostop string `xml:"disable-autostop"`
        OpenedPort string `xml:"opened-port"`
        ExternalIpAddress string `xml:"external-ip-address"`
        InternalIpAddress string `xml:"internal-ip-address"`
}

func pseudo_uuid() (uuid string) {

    b := make([]byte, 16)
    _, err := rand.Read(b)
    if err != nil {
        fmt.Println("Error: ", err)
        return
    }

    uuid = fmt.Sprintf("%X-%X-%X-%X-%X", b[0:4], b[4:6], b[6:8], b[8:10], b[10:])

    return
}

func (c *CloudBrokerClient) CreateInstance(instance *InstancePost) error {
        httpClient := &http.Client{}

        //Esempio: fare una POST per creare una istanza:

	if instance.Name=="" {
		instance.Name = pseudo_uuid()
	}
        inst, err := xml.Marshal(instance)
        if err != nil {
                log.Fatal(err)
        }
        instStr := string(inst)
        log.Printf(instStr)

        req, err := http.NewRequest("POST", "https://decide-prototype.cloudbroker.com/instances.xml", strings.NewReader(instStr))
        req.Header.Add("Content-Type", "text/xml; charset=utf-8")
        req.SetBasicAuth(c.Username, c.Password)

        resp, err := httpClient.Do(req)
        if err != nil{
                log.Fatal(err)
        }

        bodyText, err := ioutil.ReadAll(resp.Body)

        str := string(bodyText)
        log.Printf(str)

        xml.Unmarshal([]byte(str), &instance)
        log.Printf("New instance Id: " + instance.Id)
        log.Printf("New instance name: " + instance.Name)

	return nil
}

func (c *CloudBrokerClient) ReadInstance(id string, instance *InstanceGet) error {
        httpClient := &http.Client{}

        log.Printf("[INFO] Reading data for instance: %s", id)

        req, err := http.NewRequest("GET", "https://decide-prototype.cloudbroker.com/instances/"+id+".xml", nil)
        req.Header.Add("Content-Type", "text/xml; charset=utf-8")
        req.SetBasicAuth(c.Username, c.Password)
        resp, err := httpClient.Do(req)
        if err != nil{
                log.Fatal(err)
        }

        bodyText, err := ioutil.ReadAll(resp.Body)

        str := string(bodyText)
        log.Printf("[INFO] Data read for instance: " + str)

        xml.Unmarshal([]byte(str), &instance)

        log.Printf("[INFO] Instance Id: " + instance.Id)
        log.Printf("[INFO] Instance external ip: " + instance.ExternalIpAddress)
        log.Printf("[INFO] Instance internal ip: " + instance.InternalIpAddress)

        return nil
}


func (c *CloudBrokerClient) DeleteInstance(id string) error {
        httpClient := &http.Client{}

        //Esempio: fare una POST per creare una istanza:
	log.Printf("[INFO] Terminating instance: %s", id)
        req, err := http.NewRequest("PUT", "https://decide-prototype.cloudbroker.com/instances/"+id+"/stop.xml", nil)
        req.SetBasicAuth(c.Username, c.Password)
        resp, err := httpClient.Do(req)
        if err != nil{
                log.Fatal(err)
        }

        bodyText, err := ioutil.ReadAll(resp.Body)

        str := string(bodyText)
        fmt.Println(str)

        var instance InstanceGet

        xml.Unmarshal([]byte(str), &instance)
        log.Printf("[INFO] Deleted instance " + instance.Id)
        log.Printf("[INFO]     instance name: " + instance.Name)
        log.Printf("[INFO]     instance status: " + instance.Status)

        return nil
}




func main() {
	opts := plugin.ServeOpts{
		ProviderFunc: Provider,
	}
	plugin.Serve(&opts)
}

func Provider() terraform.ResourceProvider {
	return &schema.Provider{ // Source https://github.com/hashicorp/terraform/blob/v0.6.6/helper/schema/provider.go#L20-L43
		Schema:        providerSchema(),
		ResourcesMap:  providerResources(),
		ConfigureFunc: providerConfigure,
	}
}

// List of supported configuration fields for your provider.
// Here we define a linked list of all the fields that we want to
// support in our provider (api_key, endpoint, timeout & max_retries).
// More info in https://github.com/hashicorp/terraform/blob/v0.6.6/helper/schema/schema.go#L29-L142
func providerSchema() map[string]*schema.Schema {
	return map[string]*schema.Schema{
		"username": &schema.Schema{
			Type:        schema.TypeString,
			Required:    true,
			Description: "username used to authenticate with the service provider",
		},
		"password": &schema.Schema{
			Type:        schema.TypeString,
			Required:    true,
			Description: "password used to authenticate with the service provider",
		},
		"endpoint": &schema.Schema{
			Type:        schema.TypeString,
			Required:    true,
			Description: "The URL to the API",
		},
		"timeout": &schema.Schema{
			Type:        schema.TypeInt,
			Required:    true,
			Description: "Max. wait time we should wait for a successful connection to the API",
		},
		"max_retries": &schema.Schema{
			Type:        schema.TypeInt,
			Required:    true,
			Description: "The max. amount of times we will retry to connect to the API",
		},
	}
}

// List of supported resources and their configuration fields.
// Here we define da linked list of all the resources that we want to
// support in our provider. As an example, if you were to write an AWS provider
// which supported resources like ec2 instances, elastic balancers and things of that sort
// then this would be the place to declare them.
// More info here https://github.com/hashicorp/terraform/blob/v0.6.6/helper/schema/resource.go#L17-L81
func providerResources() map[string]*schema.Resource {
	return map[string]*schema.Resource{
		"cloudbroker_instance": &schema.Resource{
			SchemaVersion: 1,
			Create:        createFunc,
			Read:          readFunc,
			Update:        updateFunc,
			Delete:        deleteFunc,
			Schema: map[string]*schema.Schema{ // List of supported configuration fields for your resource
				"software_id": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
				"resource_id": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
				"region_id": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
				"instance_type_id": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
                                "vm_user": &schema.Schema{
                                        Type:     schema.TypeString,
                                        Required: true,
                                },
				"isolated": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
				"key_pair_id": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
				"disable_autostop": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
				"opened_port": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
				"name": &schema.Schema{
					Type:     schema.TypeString,
					Required: true,
				},
                                "external_ip_address": &schema.Schema{
                                        Type:     schema.TypeString,
					Computed: true,
                                },
                                "internal_ip_address": &schema.Schema{
                                        Type:     schema.TypeString,
                                        Computed: true,
                                },
			},
		},
	}
}

// This is the function used to fetch the configuration params given
// to our provider which we will use to initialise a dummy client that
// interacts with the API.
func providerConfigure(d *schema.ResourceData) (interface{}, error) {
	client := CloudBrokerClient{
		Username:     d.Get("username").(string),
		Password:     d.Get("password").(string),
		Endpoint:   d.Get("endpoint").(string),
		Timeout:    d.Get("timeout").(int),
		MaxRetries: d.Get("max_retries").(int),
	}

	// You could have some field validations here, like checking that
	// the API Key is has not expired or that the username/password
	// combination is valid, etc.

	return &client, nil
}

// The methods defined below will get called for each resource that needs to
// get created (createFunc), read (readFunc), updated (updateFunc) and deleted (deleteFunc).
// For example, if 10 resources need to be created then `createFunc`
// will get called 10 times every time with the information for the proper
// resource that is being mapped.
//
// If at some point any of these functions returns an error, Terraform will
// imply that something went wrong with the modification of the resource and it
// will prevent the execution of further calls that depend on that resource
// that failed to be created/updated/deleted.

func createFunc(d *schema.ResourceData, meta interface{}) error {
	client := meta.(*CloudBrokerClient)
	instance := InstancePost{
		SoftwareId: d.Get("software_id").(string),
		ResourceId: d.Get("resource_id").(string),
		RegionId: d.Get("region_id").(string),
		InstanceTypeId: d.Get("instance_type_id").(string),
		Isolated: d.Get("isolated").(string),
		KeyPairId: d.Get("key_pair_id").(string),
		DisableAutostop: d.Get("disable_autostop").(string),
		OpenedPort: d.Get("opened_port").(string),
		Name: d.Get("name").(string),
	}

	err := client.CreateInstance(&instance)
	if err != nil {
		return err
	}
	d.SetId(instance.Id)

	//Read new instance until the IP is assigned
	newInstance := InstanceGet{} 

	maxRetries := 90
	retries := 0
	for len(strings.TrimSpace(newInstance.ExternalIpAddress))==0 && retries<maxRetries {
		log.Printf("[INFO] Waiting to get external ip for new instance " + instance.Id)
		time.Sleep(2000 * time.Millisecond)
	        err = client.ReadInstance(d.Id(), &newInstance)
		if err != nil {
			return err
		}
		retries++
	}
	if len(strings.TrimSpace(newInstance.ExternalIpAddress))==0 {
          log.Printf("[ERROR] Ip Address not Assigned")
        }
        d.Set("external_ip_address", newInstance.ExternalIpAddress)
        d.Set("internal_ip_address", newInstance.InternalIpAddress)
	return nil
}

func readFunc(d *schema.ResourceData, meta interface{}) error {
        client := meta.(*CloudBrokerClient)

	instance := InstanceGet{}
        err := client.ReadInstance(d.Id(), &instance)
        if err != nil {
                return err
        }

        if len(strings.TrimSpace(instance.ExternalIpAddress))==0 {
		log.Printf("[ERROR] Ip Address not Assigned")
		d.SetId("")
		return nil
        }

        d.Set("external_ip_address", instance.ExternalIpAddress)
        d.Set("internal_ip_address", instance.InternalIpAddress)

	log.Printf("[INFO] From read function: external ip address -> " + d.Get("external_ip_address").(string))
	log.Printf("[INFO] From read function: internal ip address -> " + d.Get("internal_ip_address").(string))

	return nil
}

func updateFunc(d *schema.ResourceData, meta interface{}) error {
	return nil
}

func deleteFunc(d *schema.ResourceData, meta interface{}) error {
        client := meta.(*CloudBrokerClient)
        err := client.DeleteInstance(d.Id())
        if err != nil {
                return err
        }
	d.SetId("")
	return nil
}
